package hu.tigra.eing.common.entities.ingatlan;

import com.fasterxml.jackson.annotation.*;
import hu.tigra.eing.common.core.message.Request;
import hu.tigra.eing.common.core.message.Response;
import io.swagger.v3.oas.annotations.media.Schema;
import jakarta.validation.constraints.*;
import lombok.Builder;
import org.springframework.data.annotation.Immutable;

import java.util.Objects;

import static hu.tigra.eing.common.entities.ingatlan.utils.HrszComparator.DEFAULT_HRSZ_COMPARATOR;

/**
 * A földrészletek, önálló ingatlanok azonosítására szolgáló azonosító, amely település fekvésenként egyedi.
 * A helyrajzi szám felépítése:
 * - a teljes helyrajzi szám maximum 18 karakter hosszú lehet (pl.:128265/2858/BA/118)
 * - a helyrajzi számok az ingatlan típusától függően több részből állhatnak, de minimum az első rész (földrészlet azonosító) kötelező,
 * - 1. rész: maximum 6 karakter hosszan szám (földrészlet),
 * - 2. rész (opcionális): a földrészlet esetleges 'alátörés'-nek nevezett, maximum 4 karakter hosszú szám,
 * - 3. rész (opcionális): 2 karakter hosszan betű (épület). A betűk csak a magyar abc nagybetűi lehetnek. A magánhangzók közül csak az „A” betűt, a mássalhangzók közül pedig csak az egyjegyűek - a Q, W, X, Y és Z kivételével (összesen 17 betű) - használhatók. Ha a felhasználható betűk nem elegendőek, a jelölést az „AA”, „AB”, „AC”, ...,„BA”, „BB”, „BC”,... stb. jellel kezdve kell folytatni.
 * - 4. rész (opcionális): 3 karakter hosszan szám (lakás)
 * Megjelenítés szempontjából a Helyrajzi szám egy a részeiből összefűzött karaktersor, ahol az egyes részeket '/' jel választja el egymástól.
 * Kiegészítések:
 * - a földrészletek alábontása (2. rész) nem feltétlen kerül kitöltésre akkor sem ha épület vagy/és épület rész (4. rész) kitöltésre került.
 * - amennyiben egy lakás (épületrész) kerül megadásra, akkor az épület azonosító is mindig kitöltésre kerül
 */
@Builder
@Immutable
@JsonIgnoreProperties(ignoreUnknown = true)
@JsonPropertyOrder(value = {"hrszFold", "hrszAla", "hrszEp", "hrszEpresz"})
@Schema(description = "A földrészletek, önálló ingatlanok azonosítására szolgáló azonosító, amely település fekvésenként egyedi.")
public class Hrsz implements Comparable<Hrsz>, Request, Response {

    /**
     * HRSZ 1. rész: maximum 6 karakter hosszan szám (földrészlet),
     */
    @NotNull
    @Max(value = 999_999)
    @Schema($comment = "HRSZ 1. rész: maximum 6 karakter hosszan szám (földrészlet)")
    private Integer hrszFold;

    /**
     * HRSZ 2. rész: a földrészlet esetleges 'alátörés'-nek nevezett, maximum 4 karakter hosszú szám,
     */
    @Max(value = 9_999)
    @Schema($comment = "HRSZ 2. rész: a földrészlet esetleges 'alátörés'-nek nevezett, maximum 4 karakter hosszú szám")
    private Integer hrszAla;

    /**
     * HRSZ 3. rész: 2 karakter hosszan betű (épület). A betűk csak a magyar abc nagybetűi lehetnek. A magánhangzók közül csak az „A” betűt, a mássalhangzók közül pedig csak az egyjegyűek - a Q, W, X, Y és Z kivételével (összesen 17 betű) - használhatók. Ha a felhasználható betűk nem elegendőek, a jelölést az „AA”, „AB”, „AC”, ...,„BA”, „BB”, „BC”,... stb. jellel kezdve kell folytatni.
     */
    @Size(max = 2)
    @Pattern(regexp = "^[ABCDEFGHIJKLMNPRSTVYZ]{1,2}$")
    @Schema($comment = "A következő leírás lehet már elavult (EIYZ betűk is lehetnek azóta): HRSZ 3. rész: 2 karakter hosszan betű (épület). A betűk csak a magyar abc nagybetűi lehetnek. A magánhangzók közül csak az „A” betűt, a mássalhangzók közül pedig csak az egyjegyűek - a Q, W, X, Y és Z kivételével (összesen 17 betű) - használhatók. Ha a felhasználható betűk nem elegendőek, a jelölést az „AA”, „AB”, „AC”, ...,„BA”, „BB”, „BC”,... stb. jellel kezdve kell folytatni.")
    private String hrszEp;

    /**
     * HRSZ 4. rész: 3 karakter hosszan szám (lakás)
     */
    @Min(value = 1)
    @Max(value = 9_999)
    @Schema($comment = "HRSZ 4. rész: 3 karakter hosszan szám (lakás)")
    private Integer hrszEpresz;

    @JsonCreator(mode = JsonCreator.Mode.PROPERTIES)
    public Hrsz(@NotNull @Max(value = 999_999) @JsonProperty("hrszFold") Integer hrszFold,
                @Max(value = 9_999) @JsonProperty("hrszAla") Integer hrszAla,
                @Size(max = 2) @JsonProperty("hrszEp") String hrszEp,
                @Min(value = 1) @Max(value = 9_999) @JsonProperty("hrszEpresz") Integer hrszEpresz) {
        this.hrszFold = hrszFold;
        this.hrszAla = hrszAla;
        this.hrszEp = hrszEp;
        this.hrszEpresz = hrszEpresz;
    }

    public static Hrsz of(Integer hrszFold) {
        return new Hrsz(hrszFold, null, null, null);
    }

    public static Hrsz of(Integer hrszFold, Integer hrszAla) {
        return new Hrsz(hrszFold, hrszAla, null, null);
    }

    public static Hrsz of(Integer hrszFold, Integer hrszAla, String hrszEp) {
        return new Hrsz(hrszFold, hrszAla, hrszEp, null);
    }

    public static Hrsz of(Integer hrszFold, Integer hrszAla, String hrszEp, Integer hrszEpresz) {
        return new Hrsz(hrszFold, hrszAla, hrszEp, hrszEpresz);
    }

    @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
    public static Hrsz of(String hrsz) {
        return HrszParser.parse(hrsz);
    }

    /**
     * Korábbi, migrált HRSZ-ek parseolására szolgáló végpont
     * @param hrsz
     * @return
     */
    public static Hrsz ofLegacyHrsz(String hrsz) {
        return HrszParser.parse(hrsz, true);
    }

    public Integer getHrszFold() {
        return hrszFold;
    }

    public Integer getHrszAla() {
        return hrszAla;
    }

    public String getHrszEp() {
        return hrszEp;
    }

    public Integer getHrszEpresz() {
        return hrszEpresz;
    }

    @JsonIgnore
    public boolean isFoldreszlet() {
        return getHrszFold() != null && getHrszEp() == null && getHrszEpresz() == null;
    }

    @JsonIgnore
    public boolean isEpitmeny() {
        return getHrszEp() != null && getHrszEpresz() == null;
    }

    @JsonIgnore
    public boolean isEpitmenyresz() {
        return getHrszEpresz() != null;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(getHrszFold());
        if (getHrszAla() != null) {
            sb.append("/").append(getHrszAla());
        }
        if (getHrszEp() != null) {
            sb.append("/").append(getHrszEp());
        }
        if (getHrszEpresz() != null) {
            sb.append("/").append(getHrszEpresz());
        }
        return sb.toString();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Hrsz hrsz = (Hrsz) o;
        return hrszFold.equals(hrsz.hrszFold) &&
                Objects.equals(hrszAla, hrsz.hrszAla) &&
                Objects.equals(hrszEp, hrsz.hrszEp) &&
                Objects.equals(hrszEpresz, hrsz.hrszEpresz);
    }

    @Override
    public int hashCode() {
        return Objects.hash(hrszFold, hrszAla, hrszEp, hrszEpresz);
    }

    @Override
    public int compareTo(Hrsz o) {
        return DEFAULT_HRSZ_COMPARATOR.compare(this, o);
    }

    /**
     * HRSZ szöveges reprezentációjának konverziója a Hrsz formátumra.
     * Lehetséges felbontások a HRSZ-ben szerepelő '/' alapú részek számossága függvényében változik.
     * 1 rész esetén: Föld
     * 2 rész esetén: kétféle verzió lehet
     * - verzió1: Föld / Épület
     * - verzió2: Föld / Alábontás
     * 3 rész esetén: kétféle verzió lehet
     * - verzió1: Föld / Alábontás / Épület
     * - verzió2: Föld / Épület / Épületrész
     * 4 rész esetén: Föld / Alábontás / Épület / Épületrész
     * A Parser elfogadja a szöveg végi extra '/' jeleket, de a szöveg közbeni felesleges törések nem elfogadhatóak.
     */
    private static class HrszParser {
        private static final String ACCEPTABLE_EPULET_JELEK = "ABCDFGHJKLMNPRSTV";

        //TODO: 2023-09-28: Adattisztítás hiányában, ideiglenesen a régi jogszabálynak megfelelő HRSZ-ek beolvasásának engedélyezése.
        //Visszaállítása: legacyHrszValidationRules=false
        static Hrsz parse(String data) {
            return parse(data, true);
        }

        static Hrsz parse(String data, boolean legacyHrszValidationRules) {
            if (data == null || data.trim().length() == 0) {
                throw new IllegalArgumentException("ures_hrsz");
            }

            String rawData = data.trim().toUpperCase();
            String[] parts = rawData.split("/");

            int hrszFold = getHrszFold(parts[0]);
            switch (parts.length) {
                case 1:
                    return Hrsz.of(hrszFold);
                case 2:
                    if (isHrszNumeric(parts[1])) {
                        return Hrsz.of(hrszFold, getHrszAlabontas(parts[1]));
                    } else {
                        return Hrsz.of(hrszFold, null, getHrszEpuletJel(parts[1], legacyHrszValidationRules));
                    }
                case 3:
                    if (isHrszNumeric(parts[1])) {
                        return Hrsz.of(hrszFold, getHrszAlabontas(parts[1]), getHrszEpuletJel(parts[2], legacyHrszValidationRules));
                    } else {
                        return Hrsz.of(hrszFold, null, getHrszEpuletJel(parts[1], legacyHrszValidationRules), getLakas(parts[2]));
                    }
                case 4:
                    return Hrsz.of(hrszFold, getHrszAlabontas(parts[1]), getHrszEpuletJel(parts[2], legacyHrszValidationRules), getLakas(parts[3]));
                default:
                    throw new IllegalArgumentException("max_4_resz");
            }
        }

        public static boolean isHrszNumeric(String data) {
            try {
                Integer.parseInt(data);
            } catch (NumberFormatException | NullPointerException e) {
                return false;
            }
            return true;
        }

        private static String getHrszEpuletJel(String data, boolean acceptInvalidEpuletJel) {
            if (data == null) {
                throw new IllegalArgumentException("ures_epulet_jel");
            }
            int dataLength = data.length();
            if (dataLength == 0 || data.length() > 2) {
                throw new IllegalArgumentException("hibas_epuletjel");
            }

            if (!acceptInvalidEpuletJel && data.chars().anyMatch(c -> ACCEPTABLE_EPULET_JELEK.indexOf(c) == -1)) {
                throw new IllegalArgumentException("hibas_epuletjel");
            }
            return data;
        }

        private static Integer getHrszFold(String data) {
            try {
                return Integer.valueOf(data);
            } catch (NumberFormatException nfe) {
                throw new IllegalArgumentException("hibas_foldreszlet");
            }
        }

        private static Integer getHrszAlabontas(String data) {
            try {
                return Integer.valueOf(data);
            } catch (NumberFormatException nfe) {
                throw new IllegalArgumentException("hibas_foldreszlet_alabontas");
            }
        }

        private static Integer getLakas(String data) {
            try {
                return Integer.valueOf(data);
            } catch (NumberFormatException nfe) {
                throw new IllegalArgumentException("hibas_lakas");
            }
        }
    }
}
