package hu.tigra.eing.common.entities.ingatlan;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonPropertyOrder;
import hu.tigra.eing.common.core.message.Request;
import hu.tigra.eing.common.core.message.Response;
import hu.tigra.eing.common.entities.ingatlan.utils.ValidHrsz;
import hu.tigra.eing.common.entities.mdm.MDMStringReference;
import io.swagger.v3.oas.annotations.media.Schema;
import jakarta.validation.constraints.NotBlank;
import jakarta.validation.constraints.NotNull;

import java.io.Serializable;
import java.util.Objects;

import static hu.tigra.eing.common.entities.ingatlan.utils.IngatlanAzonositoComparator.DEFAULT_INGATLANAZONOSITO_COMPARATOR;

/**
 * Ingatlanok azonosításának támogatását jelentő osztály (település, fekvés és hrsz alapján)
 */
@JsonIgnoreProperties(ignoreUnknown = true)
@JsonPropertyOrder({"telepules", "fekves", "hrsz"})
@Schema(description = "Ingatlanok azonosításának támogatását jelentő osztály (település, fekvés és hrsz alapján)")
public class IngatlanAzonosito implements Serializable, Comparable<IngatlanAzonosito>, Request, Response, Cloneable {

    @NotNull
    @Schema(description = "Az ingatlan melyik településhez tartozik")
    private MDMStringReference telepules;

    @NotNull
    private Fekves fekves;

    @NotNull
    @NotBlank
    @ValidHrsz
    @Schema(description = "A földrészletek, önálló ingatlanok azonosítására szolgáló szám, amely településen azon belüli fekvésen értelmezve egyedi.", example = "128265/2858/BA/118")
    private String hrsz;

    public IngatlanAzonosito() {
    }

    public IngatlanAzonosito(@NotNull MDMStringReference telepules, @NotNull Fekves fekves, @NotNull @NotBlank @ValidHrsz String hrsz) {
        this.telepules = telepules;
        this.fekves = fekves;
        this.hrsz = hrsz;
    }

    public MDMStringReference getTelepules() {
        return telepules;
    }

    public void setTelepules(MDMStringReference telepules) {
        this.telepules = telepules;
    }

    public Fekves getFekves() {
        return fekves;
    }

    public void setFekves(Fekves fekves) {
        this.fekves = fekves;
    }

    public String getHrsz() {
        return hrsz;
    }

    public void setHrsz(String hrsz) {
        this.hrsz = hrsz;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        IngatlanAzonosito azonosito = (IngatlanAzonosito) o;
        if (telepules == null || azonosito.getTelepules() == null) return false;
        return Objects.equals(telepules.getGlobalId(), azonosito.telepules.getGlobalId()) &&
                fekves == azonosito.fekves &&
                hrsz.equals(azonosito.hrsz);
    }

    @Override
    public int hashCode() {
        return Objects.hash(telepules != null ? telepules.getGlobalId() : null, fekves, hrsz);
    }

    @Override
    public IngatlanAzonosito clone() {
        var masolat = new IngatlanAzonosito();
        masolat.setTelepules(telepules != null ? telepules.clone() : null);
        masolat.setHrsz(hrsz);
        masolat.setFekves(fekves);
        return masolat;
    }

    @Override
    public String toString() {
        return "Ingatlan{" +
                "telepules=" + telepules +
                ", fekves=" + fekves +
                ", hrsz='" + hrsz + '\'' +
                '}';
    }

    @Override
    public int compareTo(IngatlanAzonosito o) {
        return DEFAULT_INGATLANAZONOSITO_COMPARATOR.compare(this, o);
    }

    /**
     * DOM Metaadatként mentéshez való formátumra konvertáljha az osztályt.
     * @return
     */
    public String toMetadataString() {
        if (telepules == null || telepules.getGlobalId() == null || hrsz == null || fekves == null) {
            throw new IllegalStateException("Hiányos ingatlan azonosító nem sorosítható, legalátt a telepulesGlobalId, hrsz és fekves attribútumokat ki kell tölteni!");
        }
        return this.getTelepules().getGlobalId() + ";" + this.getFekves() + ";" + this.getHrsz();
    }

}
